function clamp(value, mi, ma)
	if value < mi then value = mi end
	if value > ma then value = ma end
	return value
end

function sign(x)
	return x > 0 and 1 or (x < 0 and -1 or 0)
end

function VecDist(a, b)
	return VecLength(VecSub(a, b))
end

function QuatClamp(quat, axis, min, max)
	local angles = Vec(GetQuatEuler(quat))
	angles[axis] = clamp(angles[axis], min, max)
	return QuatEuler(unpack(angles))
end

function NormalizeAngle(angle)
	if angle > 180 then
		angle = angle - 360
	elseif angle <= -180 then
		angle = angle + 360
	end
	return angle
end

shells = {}

function ShootRocket(position, direction, timer, target, target_shape)
	local shell = {}
	shell.timer = timer
	shell.max_timer = timer
	shell.pos = position
	shell.target = target
	shell.target_shape = target_shape
	shell.velocity = VecScale(direction, 50)
	shell.missile = SpawnMissile(Transform(position, QuatLookAt(position, VecAdd(position, direction))))
	table.insert(shells, shell)
end

function RocketOperation(shell_index, dt)
	local shell = shells[shell_index]
	shell.timer = shell.timer - dt
	shell.velocity = VecAdd(shell.velocity, VecScale(Vec(0, -9.8, 0), dt))
	local target_pos = shell.target

	-- Terminal guidance
	if shell.timer < shell.max_timer / 4 then
		-- Slow down timer since the missile may also slow down
		shell.timer = shell.timer + 3 * dt / 4
		if shell.target_shape ~= 0 then
			target_pos = GetExpectedShapePosition(shell.target_shape, dt)
		end
		local to_target = VecNormalize(VecSub(target_pos, shell.pos))
		shell.velocity = VecAdd(VecScale(shell.velocity, 0.95), VecScale(to_target, 2))
	end

	local next_pos = VecAdd(shell.pos, VecScale(shell.velocity, dt))
	SetBodyTransform(shell.missile, Transform(next_pos, QuatLookAt(shell.pos, next_pos)))
	SetBodyVelocity(shell.missile, shell.velocity)

	local dir = VecNormalize(VecSub(next_pos, shell.pos))
	QueryRejectBody(aircraft_body)
	QueryRequire("physical large")
	local collision_hit, hit_dist, _, hit_shape = QueryRaycast(shell.pos, dir, VecDist(next_pos, shell.pos))
	local proximity_hit = VecDist(shell.pos, target_pos) < 1
	if collision_hit or proximity_hit or shell.timer <= 0 then
		local hit_pos = VecAdd(shell.pos, VecScale(dir, hit_dist))
		if proximity_hit then
			Explosion(hit_pos, 3)
		elseif hit_shape == shell.target_shape then
			Explosion(hit_pos, 2)
		else
			Explosion(hit_pos, 1)
		end
		Delete(shell.missile)
		table.remove(shells, shell_index)
	end

	SpawnParticle("fire", shell.pos, Vec(), 0.5, 0.25)
	SpawnParticle("smoke", shell.pos, Vec(), 1, 0.75)
	shell.pos = next_pos
end

function GetExpectedShapePosition(target_shape, time)
	local min, max = GetShapeBounds(target_shape)
	local shape_center = VecLerp(min, max, 0.5)
	local position = shape_center
	local shape_hit, target_point = GetShapeClosestPoint(target_shape, shape_center)
	if shape_hit then
		position = target_point
	end
	local body = GetShapeBody(target_shape)
	local velocity = Vec(0, 0, 0)
	if IsBodyDynamic(body) then
		velocity = GetBodyVelocity(body)
	end
	return VecAdd(position, VecScale(velocity, time))
end

function SpawnMissile(transform)
	local missile_xml = "<vox tags='unbreakable nocull' texture='1 1' density='10' collide='false' prop='true' file='MOD/vox/props/missile.vox'/>"
	local spawned = Spawn(missile_xml, transform)
	local missile_body = 0
	for i = 1, #spawned do
		if GetEntityType(spawned[i]) == "body" then
			missile_body = spawned[i]
		end
	end
	return missile_body
end

function init()
	helicopter = FindVehicle("helicopter")
	aircraft_body = FindBody("body")
	--main_rotor = FindBody("mainrotor")
	--tail_rotor = FindBody("tailrotor")

	main_rotor_joint = FindJoint("mainrotor_joint")
	tail_rotor_joint = FindJoint("tailrotor_joint")

	aircraft_tr = GetBodyTransform(aircraft_body)
	--main_rotor_local_tr = TransformToLocalTransform(aircraft_tr, GetBodyTransform(main_rotor))
	--tail_rotor_local_tr = TransformToLocalTransform(aircraft_tr, GetBodyTransform(tail_rotor))

	gun_base = FindBody("base")
	gun_turret = FindBody("turret")
	joint_h = FindJoint("gun_h")
	joint_v = FindJoint("gun_v")
	local gun_base_tr = GetBodyTransform(gun_base)

	guns = {}
	local gun_locs = FindLocations("gun")
	for i = 1, #gun_locs do
		local gun_tr = GetLocationTransform(gun_locs[i])
		local gun_relative_pos = TransformToLocalPoint(gun_base_tr, gun_tr.pos)
		table.insert(guns, gun_relative_pos)
	end

	rockets = {}
	local rocket_locs = FindLocations("rocket")
	for i = 1, #rocket_locs do
		local launcher_tr = GetLocationTransform(rocket_locs[i])
		local launcher_relative_pos = TransformToLocalPoint(aircraft_tr, launcher_tr.pos)
		table.insert(rockets, launcher_relative_pos)
	end

	missiles = {}
	local missile_locs = FindLocations("missile")
	for i = 1, #missile_locs do
		local launcher_tr = GetLocationTransform(missile_locs[i])
		local launcher_relative_pos = TransformToLocalPoint(aircraft_tr, launcher_tr.pos)
		table.insert(missiles, launcher_relative_pos)
	end

	--local gun_base_tr = GetBodyTransform(gun_base)
	--gun_base_local_tr = TransformToLocalTransform(aircraft_tr, gun_base_tr)
	-- Turret transform relative to the base of the gun
	--gun_turret_local_tr = TransformToLocalTransform(gun_base_tr, GetBodyTransform(gun_turret))

	local azimut_angle = GetJointMovement(joint_h)
	azimut_min, azimut_max = GetJointLimits(joint_h)
	SetJointMotorTarget(joint_h, azimut_angle, 1, 1000)

	local elevation_angle = GetJointMovement(joint_v)
	elevation_min, elevation_max = GetJointLimits(joint_v)
	SetJointMotorTarget(joint_v, elevation_angle, 1, 1000)

	mass = GetBodyMass(aircraft_body)
	sounds = {
		chopper_l = LoadLoop("chopper-loop.ogg"),
		shoot = LoadSound("chopper-shoot0.ogg"),
		rocket = LoadSound("tools/launcher0.ogg"),
		music = "MOD/katyusha.ogg",
	}

	-- Constants
	HOVER_ALTITUDE = 0.7
	reticle1 = LoadSprite("gfx/reticle1.png")
	reticle2 = LoadSprite("gfx/reticle2.png")
	reticle3 = LoadSprite("gfx/reticle3.png")
	reticle4 = LoadSprite("gfx/reticle4.png")

	-- Weapons
	gun_reload_time = 0
	rocket_reload_time = 0
	missile_reload_time = 0
	gun_index = 1
	rocket_index = 1
	missile_index = 1

	-- Controls
	pitch = 0
	yaw = 0
	roll = 0
	collective = aircraft_tr.pos[2] + HOVER_ALTITUDE
	hover = false
	active = false

	-- Animation
	rotation_angle = 0
	coaxial_rotor = false
	new_aircraft_tr = TransformCopy(aircraft_tr)
	startup_timer = 5

	-- Camera
	zoom = 20
	camera_tr = nil
	init_camera = true
end

-- Return the signed angle between two vectors projected on the XZ plane
function GetHorizontalAngle2(forward, to_target)
	local vec1 = Vec(forward[1], 0, forward[3])
	local vec2 = Vec(to_target[1], 0, to_target[3])
	local denom = VecLength(vec1) * VecLength(vec2)
	local coef = VecDot(vec1, vec2) / denom
	if coef > 1 then
		coef = 1
	end
	local angle = math.deg(math.acos(coef))
	if VecCross(vec2, vec1)[2] <= 0 then
		return angle
	else
		return -angle
	end
end

function SetCamera()
	if camera_tr ~= nil and not GetBool("game.map.enabled") then
		SetCameraTransform(camera_tr)
		SetBool("game.vehicle.interactive", true)
	end
end

function tick(dt)
	for k in pairs(shells) do
		RocketOperation(k, dt)
	end

	if GetPlayerVehicle() ~= helicopter then
		if active then
			active = false
			-- Player leaves the helicopter
			SetJointMotor(main_rotor_joint, 0.1, 1000)
			SetJointMotor(tail_rotor_joint, 0.1, 1000)
			startup_timer = 5
			RemoveTag(helicopter, "passive")
			--StopMusic()

		end
		return
	end

	if GetVehicleHealth(helicopter) < 0.1 then
		if active then
			active = false
			-- Helicopter is too damaged to fly
			SetJointMotor(main_rotor_joint, 0.1, 1000)
			SetJointMotor(tail_rotor_joint, 0.1, 1000)
			--StopMusic()

		end
		return
	end

	if not active then
		active = true
		-- Player enters the helicopter
		aircraft_tr = GetBodyTransform(aircraft_body)
		pitch = 0
		_, yaw = GetQuatEuler(aircraft_tr.rot)
		roll = 0
		collective = aircraft_tr.pos[2] + HOVER_ALTITUDE
		--PlayMusic(sounds.music)
	end

	PlayLoop(sounds.chopper_l, aircraft_tr.pos, 10 - 2 * startup_timer)
	aircraft_tr = GetBodyTransform(aircraft_body) -- DELETE ME!
	SetCamera()

	if startup_timer > 0 then
		return
	else
		SetTag(helicopter, "passive")
	end

	if InputDown("w") then
		collective = collective + 0.1
	end
	if InputDown("s") and collective > 0 then
		collective = collective - 0.1
	end
	if InputDown("a") and roll < 30 then
		roll = roll + 0.5
	end
	if InputDown("d") and roll > -30 then
		roll = roll - 0.5
	end
	if InputPressed("h") then
		hover = not hover
	end

	local lock_yaw = yaw
	local x, y = GetQuatEuler(GetCameraTransform().rot)

	if x < pitch and pitch > -45 then
		pitch = pitch - 0.5
	end
	if x > pitch and pitch < 15 then
		pitch = pitch + 0.5
	end

	if y < yaw and not (y < yaw - 180) then
		yaw = NormalizeAngle(yaw - 1.2)
	elseif y < yaw - 180 then
		yaw = NormalizeAngle(yaw + 1.2)
	end

	if y > yaw and not (y > yaw + 180) then
		yaw = NormalizeAngle(yaw + 1.2)
	elseif y > yaw + 180 then
		yaw = NormalizeAngle(yaw - 1.2)
	end

	if hover then
		pitch = pitch - sign(pitch)
		yaw = lock_yaw
	end

	local move_dir = Vec(-dt * roll, 0, 1.5 * dt * pitch)
	local pos = TransformToParentPoint(aircraft_tr, move_dir)
	local dir = TransformToParentVec(aircraft_tr, Vec(0, -1, 0))
	local max_dist = 0.5 + math.abs(pitch / 5) + math.abs(roll / 10)
	QueryRequire("physical large")
	QueryRejectVehicle(helicopter)
	local obstacle_hit, obstacle_dist = QueryRaycast(aircraft_tr.pos, dir, max_dist)
	if obstacle_hit then
		collective = collective + (max_dist - obstacle_dist) / (2 * max_dist)
	end
	pos[2] = collective

	rotation_angle = rotation_angle + 0.6
	local rot = QuatEuler(math.sin(rotation_angle * 0.053) * 5 + pitch, math.sin(rotation_angle * 0.04) * 5 + yaw, roll)
	new_aircraft_tr = Transform(pos, rot)


	local camera_tr = GetCameraTransform()
	local camera_fwd = TransformToParentVec(camera_tr, Vec(0, 0, -1))
	dist = 50
	QueryRequire("physical large")
	QueryRejectVehicle(helicopter)
	local _, dist, _, target_shape = QueryRaycast(camera_tr.pos, camera_fwd, 100)
	local target_pos = VecAdd(camera_tr.pos, VecScale(camera_fwd, dist))
	local gun_base_tr = GetBodyTransform(gun_turret)
	local gun_fwd = TransformToParentVec(gun_base_tr, Vec(0, 0, -1))
	local world_muzzle_pos = TransformToParentPoint(gun_base_tr, guns[gun_index])
	local to_target = VecNormalize(VecSub(target_pos, world_muzzle_pos))

	-- Aim gun
	local angle_h = GetHorizontalAngle2(gun_fwd, to_target)
	local azimut_angle = GetJointMovement(joint_h) - angle_h
	azimut_angle = clamp(azimut_angle, azimut_min, azimut_max)
	SetJointMotorTarget(joint_h, azimut_angle, 1, 1000)

	local angle_v = math.deg(math.asin(to_target[2]))
	local elevation_angle = angle_v -- GetJointMovement(joint_v)
	elevation_angle = clamp(elevation_angle, elevation_min, elevation_max)
	SetJointMotorTarget(joint_v, elevation_angle, 1, 1000)

	-- Draw gun reticle
	dist = 25
	QueryRequire("physical large")
	QueryRejectVehicle(helicopter)
	_, dist = QueryRaycast(gun_base_tr.pos, gun_fwd, 100)
	DrawReticle(VecAdd(gun_base_tr.pos, VecScale(gun_fwd, dist)))

	-- Draw rocket reticle
	local aircraft_fwd = TransformToParentVec(aircraft_tr, Vec(0, 0, -1))
	dist = 25
	QueryRequire("physical large")
	QueryRejectVehicle(helicopter)
	_, dist = QueryRaycast(aircraft_tr.pos, aircraft_fwd, 100)
	DrawReticleFixed(VecAdd(aircraft_tr.pos, VecScale(aircraft_fwd, dist)), dt)

	if gun_reload_time > 0 then
		gun_reload_time = gun_reload_time - dt
	elseif InputDown("lmb") then
		gun_reload_time = 0.075
		local gun_pos = TransformToParentPoint(gun_base_tr, guns[gun_index])
		local fire_dir = TransformToParentVec(gun_base_tr, Vec(0, 0, -1))
		PlaySound(sounds.shoot, gun_pos, 5)
		Shoot(gun_pos, fire_dir, 0, 0.5)
		gun_index = gun_index % #guns + 1
	end

	if rocket_reload_time > 0 then
		rocket_reload_time = rocket_reload_time - dt
	elseif InputDown("rmb") then
		rocket_reload_time = 0.2
		local launcher_pos = TransformToParentPoint(aircraft_tr, rockets[rocket_index])
		local fire_dir = TransformToParentVec(aircraft_tr, Vec(0, 0, -1))
		PlaySound(sounds.rocket, launcher_pos, 5)
		--Shoot(launcher_pos, fire_dir, 1, 0.5)
		ShootRocket(launcher_pos, fire_dir, 5, target_pos, 0)
		rocket_index = rocket_index % #rockets + 1
	end

	if missile_reload_time > 0 then
		missile_reload_time = missile_reload_time - dt
	elseif InputPressed("mmb") then
		missile_reload_time = 1
		local launcher_pos = TransformToParentPoint(aircraft_tr, missiles[missile_index])
		local fire_dir = TransformToParentVec(aircraft_tr, Vec(0, 0, -1))
		PlaySound(sounds.rocket, launcher_pos, 5)
		--Shoot(launcher_pos, fire_dir, 1)
		ShootRocket(launcher_pos, fire_dir, 5, target_pos, target_shape)
		missile_index = missile_index % #missiles + 1
	end

	ConstrainPosition(aircraft_body, 0, aircraft_tr.pos, new_aircraft_tr.pos)
	ConstrainOrientation(aircraft_body, 0, aircraft_tr.rot, new_aircraft_tr.rot)
end

function DrawReticle(target_pos)
	local t = Transform(target_pos, QuatLookAt(target_pos, GetCameraTransform().pos))
	DrawSprite(reticle4, t, 3, 1.5, 0.5, 0, 0, 1, false, false)
	DrawSprite(reticle4, t, 3, 1.5, 0.5, 0, 0, 1, true, false)
end

function DrawReticleFixed(target_pos, dt)
	local t = Transform(target_pos, QuatLookAt(target_pos, aircraft_tr.pos))
	local tr = QuatEuler(0, 0, dt * 60)

	t.rot = QuatRotateQuat(t.rot, tr)
	DrawSprite(reticle1, t, 1.2, 1.2, 0.5, 0, 0, 1, false, false)
	DrawSprite(reticle1, t, 1.2, 1.2, 0.5, 0, 0, 1, true, false)

	tr = QuatEuler(0, 0, dt * -80)
	t.rot = QuatRotateQuat(t.rot, tr)
	DrawSprite(reticle2, t, 1.2, 1.2, 0.5, 0, 0, 1, false, false)
	DrawSprite(reticle2, t, 1.2, 1.2, 0.5, 0, 0, 1, true, false)

	tr = QuatEuler(0, 0, dt * 100)
	t.rot = QuatRotateQuat(t.rot, tr)
	DrawSprite(reticle3, t, 1.2, 1.2, 0.5, 0, 0, 1, false, false)
	DrawSprite(reticle3, t, 1.2, 1.2, 0.5, 0, 0, 1, true, false)
end

function update(dt)
	if active and GetPlayerVehicle() == helicopter then
		-- Get camera center
		local driver_pos = GetVehicleDriverPos(helicopter)
		local vehicle_tr = GetVehicleTransform(helicopter)
		local world_driver_pos = TransformToParentPoint(vehicle_tr, driver_pos)

		-- Get camera zoom level
		local input = InputValue("mousewheel")
		zoom = clamp(zoom - input * 0.5, 0, 30)
		local first_person_camera = zoom < 12
		if first_person_camera and -input > 0 then
			zoom = 12
			first_person_camera = false
		elseif first_person_camera and -input < 0 then
			zoom = 0
		end

		-- Reset camera rotation
		if init_camera then
			--camera_rot = GetCameraTransform().rot
			camera_rot = aircraft_tr.rot
			init_camera = false
		end

		-- Get camera rotation
		camera_rot = QuatRotateQuat(camera_rot, QuatEuler(-InputValue("cameray") * 120, -InputValue("camerax") * 120, 0))
		if first_person_camera then
			camera_rot = QuatClamp(camera_rot, 1, -80, 80)
		else
			camera_rot = QuatClamp(camera_rot, 1, -50, 30)
		end

		-- Get vehicle forward speed
		local vehicle_main_body = GetVehicleBody(vehicle)
		local vel = TransformToLocalVec(vehicle_tr, GetBodyVelocity(vehicle_main_body))
		local speed = -vel[3]

		-- Get camera position
		local camera_center_tr = Transform(world_driver_pos, camera_rot)
		local camera_pos = TransformToParentPoint(camera_center_tr, Vec(0, 0, zoom))
		-- Compensate for speed
		camera_pos = VecAdd(camera_pos, QuatRotateVec(vehicle_tr.rot, Vec(0, 0, -speed * dt)))

		-- Set camera
		camera_rot = QuatClamp(camera_rot, 3, 0, 0)
		camera_tr = Transform(camera_pos, camera_rot)
		SetCamera()

		-- Set aircraft transform
		if startup_timer > 0 then
			startup_timer = startup_timer - dt
		else
			ConstrainPosition(aircraft_body, 0, aircraft_tr.pos, new_aircraft_tr.pos)
			ConstrainOrientation(aircraft_body, 0, aircraft_tr.rot, new_aircraft_tr.rot)
		end
		SetJointMotor(main_rotor_joint, 15 - startup_timer * 3)
		SetJointMotor(tail_rotor_joint, 15 - startup_timer * 3)
		aircraft_tr = TransformCopy(new_aircraft_tr)
	else
		init_camera = true
		camera_tr = nil
	end
end

function DrawPoint(point, name)
	local x, y, d = UiWorldToPixel(point)
	if d > 0 then
		UiTranslate(x, y)
		UiAlign("center middle")
		UiImage("ui/common/dot.png")
		UiTranslate(0, 20)
		UiFont("regular.ttf", 20)
		UiText(name)
	end
end

function DebugDraw()
	local aircraft_tr = GetBodyTransform(aircraft_body)
	local gun_base_tr = GetBodyTransform(gun_base)
	for i = 1, #missiles do
		local launcher_pos = TransformToParentPoint(aircraft_tr, missiles[i])
		DrawPoint(launcher_pos, "Missile " .. i)
	end
	for i = 1, #rockets do
		local launcher_pos = TransformToParentPoint(aircraft_tr, rockets[i])
		DrawPoint(launcher_pos, "Rocket " .. i)
	end
	for i = 1, #guns do
		local gun_pos = TransformToParentPoint(gun_base_tr, guns[i])
		DrawPoint(gun_pos, "Gun " .. i)
	end
end
